/************************************************************************************************
*  Filename: duplicatefinder.c
*  Author: <kasskloff AT gmail.com>
*  Description: A program to scan directories for duplicate files and optionally delete them
*  
*  Permission is hereby granted, free of charge, to any person obtaining a copy
*  of this software and associated documentation files (the "Software"), to deal
*  in the Software without restriction, including without limitation the rights
*  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
*  copies of the Software, and to permit persons to whom the Software is
*  furnished to do so, subject to the following conditions:
*  
*  The above copyright notice and this permission notice shall be included in all
*  copies or substantial portions of the Software.
*  
*  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
*  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
*  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
*  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
*  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
*  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
*  SOFTWARE.
************************************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <windows.h>
#include <time.h>
#include "md5hash.h"

#define STRING_EMPTY ""
#define STRING_WEMPTY L""
#define INITIAL_FILEARRAY_SIZE 100
#define FILE_CHUNK_SIZE 512000
#define RECURSION_RECURSION 1
#define PROMPT_DELETE 1
#define EXIT_EXIT 1
#define EXIT_NO_EXIT 0

#define ERROR_NO_ERROR 0x00
#define ERROR_TOO_LONG 0x01
#define ERROR_NO_OPEN 0x02
#define ERROR_INVALID_PATH 0x03
#define ERROR_NO_SIZE 0x04
#define ERROR_NO_DELETE 0x05
#define ERROR_NO_ALLOC 0x06

typedef struct
{
	char* fileInfo;
	unsigned char* hash;
} fileArrayEntryA;

typedef struct
{
	fileArrayEntryA* array;
	size_t elements;
	size_t size;
} fileArrayA;

typedef struct
{
	wchar_t* fileInfo;
	unsigned char* hash;
} fileArrayEntryW;

typedef struct
{
	fileArrayEntryW* array;
	size_t elements;
	size_t size;
} fileArrayW;

typedef struct
{
	char** array;
	size_t elements;
	size_t size;
} pathArrayA;

typedef struct
{
	wchar_t** array;
	size_t elements;
	size_t size;
} pathArrayW;

void RunProgramA(int argc, char* argv[]);
void RunProgramW(int argc, wchar_t* argv[]);
void SearchFilesA(fileArrayA* files, const char* directory, int recursion, int prompt);
void SearchFilesW(fileArrayW* files, const wchar_t* directory, int recursion, int prompt);
void AddFileA(fileArrayA* files, fileArrayEntryA* entry);
void AddFileW(fileArrayW* files, fileArrayEntryW* entry);
void RemoveFileA(fileArrayA* files, size_t index);
void RemoveFileW(fileArrayW* files, size_t index);
void AddPathA(fileArrayA* files, pathArrayA* paths, char* path);
void AddPathW(fileArrayW* files, pathArrayW* paths, wchar_t* path);
void FreePathArrayA(pathArrayA* paths);
void FreePathArrayW(pathArrayW* paths);
char* GetErrorMessageA(DWORD errNo);
wchar_t* GetErrorMessageW(DWORD errNo);
int GetConfirmation();
void DisplayErrorA(fileArrayA* files, DWORD errNo, int errType, const char* file, int doExit);
void DisplayErrorW(fileArrayW* files, DWORD errNo, int errType, const wchar_t* file, int doExit);
void ExitA(fileArrayA* files, DWORD errNo);
void ExitW(fileArrayW* files, DWORD errNo);

int main(int argc, char* argv[])
{
	SetConsoleTitle("Duplicate Finder");
	
	typedef BOOL WINAPI (*import_Wow64DisableWow64FsRedirection)(PVOID* OldValue);
	HINSTANCE kernel32;
	import_Wow64DisableWow64FsRedirection Wow64DisableWow64FsRedirection;
	kernel32 = LoadLibrary("kernel32.dll");
	Wow64DisableWow64FsRedirection = (import_Wow64DisableWow64FsRedirection)GetProcAddress(kernel32, "Wow64DisableWow64FsRedirection");
	if (Wow64DisableWow64FsRedirection)
	{
		PVOID oldVal;
		Wow64DisableWow64FsRedirection(&oldVal);
	}
	
	OSVERSIONINFO os;
	memset(&os, 0, sizeof(OSVERSIONINFO));
	os.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&os);
	
	DWORD unicodeSupport = (os.dwMajorVersion > 4 ? 1 : 0);
	if (unicodeSupport)
	{
		int argcW;
		wchar_t** argvW;
		argvW = (wchar_t**)CommandLineToArgvW(GetCommandLineW(), &argcW);
		if (argvW != NULL)
			RunProgramW(argcW, argvW);
	}
	RunProgramA(argc, argv);
}

void RunProgramA(int argc, char* argv[])
{
	fileArrayA files; 
	files.array = NULL;
	
	pathArrayA searchList;
	searchList.size = 5;
	searchList.elements = 0;
	searchList.array = calloc(searchList.size, sizeof(char*));
	
	char input[MAX_PATH];
	memset(input, '\0', sizeof(input));
	if (argc > 1)
	{
		if (strlen(argv[1]) > MAX_PATH)
			DisplayErrorA(&files, EXIT_FAILURE, ERROR_TOO_LONG, argv[1], EXIT_EXIT);

		strcpy(input, argv[1]);
		AddPathA(&files, &searchList, input);
	}
	else
	{
		int done = 0;
		printf("Enter the path(s) you wish to search, one per line, and hit the return when done:\n");
		do
		{
			memset(input, '\0', sizeof(input));
			printf("> ");
			fgets(input, sizeof(input), stdin);
			if (input[0] == '\0' || input[0] == '\r' || input[0] == '\n')
			{
				if (searchList.elements == 0)
					ExitA(&files, EXIT_SUCCESS);
				else
				{
					done = 1;
					break;
				}
			}
			if (input[strlen(input) - 1] == '\n')
				input[strlen(input) - 1] = '\0';
			
			AddPathA(&files, &searchList, input);
		} while (!done);
	}
	
	int recursion = -1;
	int prompt = -1;
	
	if (argc > 2)
	{
		for (int i = 2; i < argc; i++)
		{
			char* tempArg = calloc(strlen(argv[i]) + 1, sizeof(char));
			if (tempArg == NULL)
				DisplayErrorA(&files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
			
			strcpy(tempArg, argv[i]);
			strlwr(tempArg);

			if (!strcmp(tempArg, "/r") || !strcmp(tempArg, "/recursive"))
				recursion = 1;
			else if (!strcmp(tempArg, "/nr") || !strcmp(tempArg, "/norecursion"))
				recursion = 0;
			else if (!strcmp(tempArg, "/p") || !strcmp(tempArg, "/prompt"))
				prompt = 1;
			else if (!strcmp(tempArg, "/np") || !strcmp(tempArg, "/noprompt"))
				prompt = 0;
			else
				AddPathA(&files, &searchList, argv[i]);
			
			free(tempArg);
		}
	}
	
	for (size_t i = 0; i < searchList.elements; i++)
	{
		DWORD attrib = GetFileAttributes(searchList.array[i]);
		if (attrib == INVALID_FILE_ATTRIBUTES)
		{
			DisplayErrorA(&files, GetLastError(), ERROR_NO_OPEN, searchList.array[i], EXIT_NO_EXIT);
			FreePathArrayA(&searchList);
			ExitA(&files, GetLastError());
		}
		else if (!(attrib & FILE_ATTRIBUTE_DIRECTORY))
		{
			DisplayErrorA(&files, EXIT_FAILURE, ERROR_INVALID_PATH, searchList.array[i], EXIT_EXIT);
			FreePathArrayA(&searchList);
			ExitA(&files, EXIT_FAILURE);
		}
	}
	
	if (recursion == -1)
	{
		printf("\nWould you like to recursively search all subdirectories? (Y/N): ");
		recursion = GetConfirmation();
	}
	
	if (prompt == -1)
	{
		printf("\nWould you like to be asked whether or not to delete duplicate files? (Y/N): ");
		prompt = GetConfirmation();
	}
	
	files.array = calloc(INITIAL_FILEARRAY_SIZE, sizeof(fileArrayEntryA));
	files.elements = 0;
	files.size = INITIAL_FILEARRAY_SIZE;

	SetConsoleTitle("Duplicate Finder - Searching...");
	printf("Searching...\n");
	double start = ((float)clock() / CLOCKS_PER_SEC);
	for (size_t i = 0; i < searchList.elements; i++)
		SearchFilesA(&files, searchList.array[i], recursion, prompt);
	double fin = ((float)clock() / CLOCKS_PER_SEC);
	SetConsoleTitle("Duplicate Finder");
	printf("\nDone! (%f seconds)\n", (fin - start));
	
	FreePathArrayA(&searchList);
	ExitA(&files, EXIT_SUCCESS);
}

void RunProgramW(int argc, wchar_t* argv[])
{
	fflush(stdout);
	_setmode(_fileno(stdout), 0x20000);
	fileArrayW files; 
	files.array = NULL;
	
	pathArrayW searchList;
	searchList.size = 5;
	searchList.elements = 0;
	searchList.array = calloc(searchList.size, sizeof(wchar_t*));
	
	wchar_t input[MAX_PATH];
	memset(input, '\0', sizeof(input));
	if (argc > 1)
	{
		if (wcslen(argv[1]) > MAX_PATH)
			DisplayErrorW(&files, EXIT_FAILURE, ERROR_TOO_LONG, argv[1], EXIT_EXIT);

		wcscpy(input, argv[1]);
		AddPathW(&files, &searchList, input);
	}
	else
	{
		int done = 0;
		wprintf(L"Enter the path(s) you wish to search, one per line, and hit the return when done:\n");
		do
		{
			memset(input, '\0', sizeof(input));
			wprintf(L"> ");
			fgetws(input, sizeof(input), stdin);
			if (input[0] == '\0' || input[0] == '\r' || input[0] == '\n')
			{
				if (searchList.elements == 0)
					ExitW(&files, EXIT_SUCCESS);
				else
				{
					done = 1;
					break;
				}
			}
			if (input[wcslen(input) - 1] == '\n')
				input[wcslen(input) - 1] = '\0';
			
			AddPathW(&files, &searchList, input);
		} while (!done);
	}
	
	int recursion = -1;
	int prompt = -1;
	
	if (argc > 2)
	{
		for (int i = 2; i < argc; i++)
		{
			wchar_t* tempArg = calloc(wcslen(argv[i]) + 1, sizeof(wchar_t));
			if (tempArg == NULL)
				DisplayErrorW(&files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_WEMPTY, EXIT_EXIT);
			
			wcscpy(tempArg, argv[i]);
			wcslwr(tempArg);

			if (!wcscmp(tempArg, L"/r") || !wcscmp(tempArg, L"/recursive"))
				recursion = 1;
			else if (!wcscmp(tempArg, L"/nr") || !wcscmp(tempArg, L"/norecursion"))
				recursion = 0;
			else if (!wcscmp(tempArg, L"/p") || !wcscmp(tempArg, L"/prompt"))
				prompt = 1;
			else if (!wcscmp(tempArg, L"/np") || !wcscmp(tempArg, L"/noprompt"))
				prompt = 0;
			else
				AddPathW(&files, &searchList, argv[i]);
			
			free(tempArg);
		}
	}
	LocalFree((LPWSTR*)argv);
	
	for (size_t i = 0; i < searchList.elements; i++)
	{
		DWORD attrib = GetFileAttributesW(searchList.array[i]);
		if (attrib == INVALID_FILE_ATTRIBUTES)
		{
			DisplayErrorW(&files, GetLastError(), ERROR_NO_OPEN, searchList.array[i], EXIT_NO_EXIT);
			FreePathArrayW(&searchList);
			ExitW(&files, GetLastError());
		}
		else if (!(attrib & FILE_ATTRIBUTE_DIRECTORY))
		{
			DisplayErrorW(&files, EXIT_FAILURE, ERROR_INVALID_PATH, searchList.array[i], EXIT_EXIT);
			FreePathArrayW(&searchList);
			ExitW(&files, EXIT_FAILURE);
		}
	}
	
	if (recursion == -1)
	{
		wprintf(L"\nWould you like to recursively search all subdirectories? (Y/N): ");
		recursion = GetConfirmation();
	}
	
	if (prompt == -1)
	{
		wprintf(L"\nWould you like to be asked whether or not to delete duplicate files? (Y/N): ");
		prompt = GetConfirmation();
	}
	
	files.array = calloc(INITIAL_FILEARRAY_SIZE, sizeof(fileArrayEntryW));
	files.elements = 0;
	files.size = INITIAL_FILEARRAY_SIZE;

	SetConsoleTitle("Duplicate Finder - Searching...");
	wprintf(L"Searching...\n");
	double start = ((float)clock() / CLOCKS_PER_SEC);
	for (size_t i = 0; i < searchList.elements; i++)
		SearchFilesW(&files, searchList.array[i], recursion, prompt);
	double fin = ((float)clock() / CLOCKS_PER_SEC);
	SetConsoleTitle("Duplicate Finder");
	wprintf(L"\nDone! (%f seconds)\n", (fin - start));
	
	FreePathArrayW(&searchList);
	ExitW(&files, EXIT_SUCCESS);
}

void SearchFilesA(fileArrayA* files, const char* directory, int recursion, int prompt)
{
	WIN32_FIND_DATA file;
	HANDLE handle;
	
	MD5 md5;
	unsigned char hash[16];
	
	char pattern[MAX_PATH];
	memset(pattern, '\0', sizeof(pattern));
	strcat(pattern, directory);
	if (pattern[strlen(pattern) - 1] != '\\')
		strcat(pattern, "\\*.*");
	else
		strcat(pattern, "*.*");
	
	char* title = calloc(strlen(directory) + 35, sizeof(char));
	if (title == NULL)
		DisplayErrorA(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
	
	sprintf(title, "Duplicate finder - Searching... [%s]", directory);
	SetConsoleTitle(title);
	free(title);
	
	handle = FindFirstFile(pattern, &file);
	if (handle == INVALID_HANDLE_VALUE)
	{
		DisplayErrorA(files, GetLastError(), ERROR_NO_OPEN, directory, EXIT_NO_EXIT);
		return;
	}
	
	do
	{
		if (strcmp(file.cFileName, ".") && strcmp(file.cFileName, ".."))
		{
			char* fullName = calloc((sizeof(directory) + sizeof(file.cFileName)) + 1, sizeof(char));
			if (fullName == NULL)
				DisplayErrorA(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
			
			strcat(fullName, directory);
			if (fullName[strlen(fullName) - 1] != '\\')
				strcat(fullName, "\\");
			strcat(fullName, file.cFileName);
			
			DWORD attrib = GetFileAttributes(fullName);
			if (attrib == INVALID_FILE_ATTRIBUTES)
			{
				DisplayErrorA(files, GetLastError(), ERROR_NO_OPEN, fullName, EXIT_NO_EXIT);
				free(fullName);
				continue;
			}
			else if (attrib & FILE_ATTRIBUTE_DIRECTORY)
			{
				if (recursion == RECURSION_RECURSION)
					SearchFilesA(files, fullName, RECURSION_RECURSION, prompt);
			}
			else
			{
				HANDLE read = CreateFile(fullName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
				DWORD fileSize = GetFileSize(read, NULL);
				if (fileSize == INVALID_FILE_SIZE)
				{
					DisplayErrorA(files, GetLastError(), ERROR_NO_SIZE, fullName, EXIT_NO_EXIT);
					free(fullName);
					CloseHandle(read);
					continue;
				}
				
				Initialise_md5(&md5);
				unsigned char* buffer = calloc(FILE_CHUNK_SIZE, sizeof(unsigned char));
				if (buffer == NULL)
				{
					free(fullName);
					CloseHandle(read);
					DisplayErrorA(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
				}
				char finalHash[33];
				memset(finalHash, '\0', sizeof(finalHash));
				DWORD totalBytes = 0;
				
				while (totalBytes < fileSize)
				{
					DWORD bytesRead = 0;
					ReadFile(read, buffer, FILE_CHUNK_SIZE, &bytesRead, NULL);
					Update_md5(&md5, buffer, bytesRead);
					totalBytes += bytesRead;
				}
				
				free(buffer);
				CloseHandle(read);
				Finalise_md5(&md5, hash);
				
				int doAddFile = 1;
				for (size_t i = 0; i < files->elements; i++)
				{
					if (files->array[i].hash == NULL)
						continue;
					
					if (!memcmp(files->array[i].hash, hash, 16))
					{
						printf("\n1: [%s] (%u bytes) \nis a duplicate of file \n2: [%s]\n", fullName, (unsigned int)fileSize, files->array[i].fileInfo);
						
						if (prompt == PROMPT_DELETE)
						{
							printf("Delete file? (1/2/N): ");
							
							char confirm[3];
							fgets(confirm, sizeof(confirm), stdin);
							if (confirm[0] == '1')
							{
								if (DeleteFile(fullName))
								{
									printf("\nFile [%s] deleted!\n", fullName);
									doAddFile = 0;
								}
								else
									DisplayErrorA(files, GetLastError(), ERROR_NO_DELETE, fullName, EXIT_NO_EXIT);
							}
							else if (confirm[0] == '2')
							{
								if (DeleteFile(files->array[i].fileInfo))
								{
									printf("\nFile [%s] deleted!\n", files->array[i].fileInfo);
									RemoveFileA(files, i);
								}
								else
									DisplayErrorA(files, GetLastError(), ERROR_NO_DELETE, fullName, EXIT_NO_EXIT);
							}
						}
						break;
					}
				}
				
				if (doAddFile)
				{
					fileArrayEntryA entry;
					entry.fileInfo = malloc((strlen(fullName) + 1) * sizeof(char));
					entry.hash = malloc(16 * sizeof(unsigned char));
					if (entry.fileInfo == NULL || entry.hash == NULL)
					{
						free(fullName);
						DisplayErrorA(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
					}
					strcpy(entry.fileInfo, fullName);
					memcpy(entry.hash, hash, 16);
					AddFileA(files, &entry);
					free(entry.fileInfo);
					free(entry.hash);
				}
			}
			free(fullName);
		}
	} while (FindNextFile(handle, &file));
	
	FindClose(handle);
}

void SearchFilesW(fileArrayW* files, const wchar_t* directory, int recursion, int prompt)
{
	WIN32_FIND_DATAW file;
	HANDLE handle;
	
	MD5 md5;
	unsigned char hash[16];
	
	wchar_t pattern[MAX_PATH];
	memset(pattern, '\0', sizeof(pattern));
	wcscat(pattern, directory);
	if (pattern[wcslen(pattern) - 1] != '\\')
		wcscat(pattern, L"\\*.*");
	else
		wcscat(pattern, L"*.*");
	
	wchar_t* title = calloc(wcslen(directory) + 35, sizeof(wchar_t));
	if (title == NULL)
		DisplayErrorW(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_WEMPTY, EXIT_EXIT);
	
	swprintf(title, L"Duplicate finder - Searching... [%s]", directory);
	SetConsoleTitleW(title);
	free(title);
	
	handle = FindFirstFileW(pattern, &file);
	if (handle == INVALID_HANDLE_VALUE)
	{
		DisplayErrorW(files, GetLastError(), ERROR_NO_OPEN, directory, EXIT_NO_EXIT);
		return;
	}
	
	do
	{
		if (wcscmp(file.cFileName, L".") && wcscmp(file.cFileName, L".."))
		{
			wchar_t* fullName = calloc((sizeof(directory) + sizeof(file.cFileName)) + 1, sizeof(wchar_t));
			if (fullName == NULL)
				DisplayErrorW(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_WEMPTY, EXIT_EXIT);
			
			wcscat(fullName, directory);
			if (fullName[wcslen(fullName) - 1] != '\\')
				wcscat(fullName, L"\\");
			wcscat(fullName, file.cFileName);
			
			DWORD attrib = GetFileAttributesW(fullName);
			if (attrib == INVALID_FILE_ATTRIBUTES)
			{
				DisplayErrorW(files, GetLastError(), ERROR_NO_OPEN, fullName, EXIT_NO_EXIT);
				free(fullName);
				continue;
			}
			else if (attrib & FILE_ATTRIBUTE_DIRECTORY)
			{
				if (recursion == RECURSION_RECURSION)
					SearchFilesW(files, fullName, RECURSION_RECURSION, prompt);
			}
			else
			{
				HANDLE read = CreateFileW(fullName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
				DWORD fileSize = GetFileSize(read, NULL);
				if (fileSize == INVALID_FILE_SIZE)
				{
					DisplayErrorW(files, GetLastError(), ERROR_NO_SIZE, fullName, EXIT_NO_EXIT);
					free(fullName);
					CloseHandle(read);
					continue;
				}
				
				Initialise_md5(&md5);
				unsigned char* buffer = malloc(FILE_CHUNK_SIZE * sizeof(unsigned char));
				if (buffer == NULL)
				{
					free(fullName);
					CloseHandle(read);
					DisplayErrorW(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_WEMPTY, EXIT_EXIT);
				}
				
				DWORD totalBytes = 0;
				while (totalBytes < fileSize)
				{
					DWORD bytesRead = 0;
					ReadFile(read, buffer, FILE_CHUNK_SIZE, &bytesRead, NULL);
					Update_md5(&md5, buffer, bytesRead);
					totalBytes += bytesRead;
				}
				
				free(buffer);
				CloseHandle(read);
				Finalise_md5(&md5, hash);
				
				int doAddFile = 1;				
				for (size_t i = 0; i < files->elements; i++)
				{
					if (files->array[i].hash == NULL)
						continue;
					
					if (!memcmp(files->array[i].hash, hash, 16))
					{
						wprintf(L"\n1: [%s] (%u bytes) \nis a duplicate of file \n2: [%s]\n", fullName, (unsigned int)fileSize, files->array[i].fileInfo);
						
						if (prompt == PROMPT_DELETE)
						{
							wprintf(L"Delete file? (1/2/N): ");
							
							wchar_t confirm[3];
							fgetws(confirm, sizeof(confirm), stdin);
							if (confirm[0] == '1')
							{
								if (DeleteFileW(fullName))
								{
									wprintf(L"\nFile [%s] deleted!\n", fullName);
									doAddFile = 0;
								}
								else
									DisplayErrorW(files, GetLastError(), ERROR_NO_DELETE, fullName, EXIT_NO_EXIT);
							}
							else if (confirm[0] == '2')
							{
								if (DeleteFileW(files->array[i].fileInfo))
								{
									wprintf(L"\nFile [%s] deleted!\n", files->array[i].fileInfo);
									RemoveFileW(files, i);
								}
								else
									DisplayErrorW(files, GetLastError(), ERROR_NO_DELETE, fullName, EXIT_NO_EXIT);
							}
						}
						break;
					}
				}
				
				if (doAddFile)
				{
					fileArrayEntryW entry;
					entry.fileInfo = malloc((wcslen(fullName) + 1) * sizeof(wchar_t));
					entry.hash = malloc(16 * sizeof(unsigned char));
					if (entry.fileInfo == NULL || entry.hash == NULL)
					{
						free(fullName);
						DisplayErrorW(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_WEMPTY, EXIT_EXIT);
					}
					wcscpy(entry.fileInfo, fullName);
					memcpy(entry.hash, hash, 16);
					AddFileW(files, &entry);
					free(entry.fileInfo);
					free(entry.hash);
				}
			}
			free(fullName);
		}
	} while (FindNextFileW(handle, &file));
	
	FindClose(handle);
}

void AddFileA(fileArrayA* files, fileArrayEntryA* entry)
{
	files->elements++;
	if (files->elements > files->size)
	{
		files->size *= 2;
		fileArrayEntryA* expand = realloc(files->array, (files->size * sizeof(fileArrayEntryA)));
		if (expand == NULL)
			DisplayErrorA(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
		
		files->array = expand;
	}
	files->array[files->elements - 1].fileInfo = malloc((strlen(entry->fileInfo) + 1) * sizeof(char));
	files->array[files->elements - 1].hash = malloc(16 * sizeof(unsigned char));
	strcpy(files->array[files->elements - 1].fileInfo, entry->fileInfo);
	memcpy(files->array[files->elements - 1].hash, entry->hash, 16);
}

void AddFileW(fileArrayW* files, fileArrayEntryW* entry)
{
	files->elements++;
	if (files->elements > files->size)
	{
		files->size *= 2;
		fileArrayEntryW* expand = realloc(files->array, (files->size * sizeof(fileArrayEntryW)));
		if (expand == NULL)
			DisplayErrorW(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_WEMPTY, EXIT_EXIT);
		
		files->array = expand;
	}
	files->array[files->elements - 1].fileInfo = malloc((wcslen(entry->fileInfo) + 1) * sizeof(wchar_t));
	files->array[files->elements - 1].hash = malloc(16 * sizeof(unsigned char));
	
	if (files->array[files->elements - 1].fileInfo == NULL || files->array[files->elements - 1].hash == NULL)
		DisplayErrorW(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_WEMPTY, EXIT_EXIT);
	
	wcscpy(files->array[files->elements - 1].fileInfo, entry->fileInfo);
	memcpy(files->array[files->elements - 1].hash, entry->hash, 16);
}

void RemoveFileA(fileArrayA* files, size_t index)
{
	free(files->array[index].fileInfo);
	free(files->array[index].hash);
	files->array[index].fileInfo = NULL;
	files->array[index].hash = NULL;
}

void RemoveFileW(fileArrayW* files, size_t index)
{
	free(files->array[index].fileInfo);
	free(files->array[index].hash);
	files->array[index].fileInfo = NULL;
	files->array[index].hash = NULL;
}

void AddPathA(fileArrayA* files, pathArrayA* paths, char* path)
{
	for (size_t i = 0; i < paths->elements; i++)
	{
		if (!_stricmp(path, paths->array[i]))
			return;
	}
	
	paths->elements++;
	if (paths->elements > paths->size)
	{
		paths->size *= 2;
		char** expand = realloc(paths->array, (paths->size * sizeof(char*)));
		if (expand == NULL)
			DisplayErrorA(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
		
		paths->array = expand;
	}
	paths->array[paths->elements - 1] = calloc(strlen(path) + 1, sizeof(char));
	
	if (paths->array[paths->elements - 1] == NULL)
		DisplayErrorA(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_EMPTY, EXIT_EXIT);
	
	strcpy(paths->array[paths->elements - 1], path);	
}

void AddPathW(fileArrayW* files, pathArrayW* paths, wchar_t* path)
{
	for (size_t i = 0; i < paths->elements; i++)
	{
		if (!_wcsicmp(path, paths->array[i]))
			return;
	}
	
	paths->elements++;
	if (paths->elements > paths->size)
	{
		paths->size *= 2;
		wchar_t** expand = realloc(paths->array, (paths->size * sizeof(wchar_t*)));
		if (expand == NULL)
			DisplayErrorW(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_WEMPTY, EXIT_EXIT);
		
		paths->array = expand;
	}
	paths->array[paths->elements - 1] = calloc(wcslen(path) + 1, sizeof(wchar_t));
	
	if (paths->array[paths->elements - 1] == NULL)
		DisplayErrorW(files, EXIT_FAILURE, ERROR_NO_ALLOC, STRING_WEMPTY, EXIT_EXIT);
	
	wcscpy(paths->array[paths->elements - 1], path);
}

void FreePathArrayA(pathArrayA* paths)
{
	for (size_t x = 0; x < paths->elements; x++)
		free(paths->array[x]);
	free(paths->array);
}

void FreePathArrayW(pathArrayW* paths)
{
	for (size_t x = 0; x < paths->elements; x++)
		free(paths->array[x]);
	free(paths->array);
}

char* GetErrorMessageA(DWORD errNo)
{
	LPVOID buffer;
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, errNo, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)&buffer, 0, NULL);
	char* message = (char*)buffer;
	if (message[strlen(message) - 2] == '\r')
		message[strlen(message) - 2] = '\0';
	
	return message;
}

wchar_t* GetErrorMessageW(DWORD errNo)
{
	LPVOID buffer;
	FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, errNo, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPWSTR)&buffer, 0, NULL);
	wchar_t* message = (wchar_t*)buffer;
	if (message[wcslen(message) - 2] == '\r')
		message[wcslen(message) - 2] = '\0';
	
	return message;
}

int GetConfirmation()
{
	char confirm[3];
	fgets(confirm, sizeof(confirm), stdin);
	return (confirm[0] == 'y' || confirm[0] == 'Y');
}

void DisplayErrorA(fileArrayA* files, DWORD errNo, int errType, const char* file, int doExit)
{
	if (errNo != ERROR_NO_ERROR)
	{
		char* errMsg;
		if (errType != ERROR_NO_ALLOC)
			errMsg = GetErrorMessageA(errNo);
		switch (errType)
		{
			case ERROR_TOO_LONG:
				printf("\nPath [%s] exceeds maximum path length of %u characters\n", file, MAX_PATH);
			break;
			
			case ERROR_NO_OPEN:
				printf("\nCannot open file/directory [%s] ((%u) %s)\n", file, (unsigned int)errNo, errMsg);
			break;
			
			case ERROR_INVALID_PATH:
				printf("\nPath [%s] is not a valid directory\n", file);
			break;
			
			case ERROR_NO_SIZE:
				printf("\nCannot get size of file [%s] ((%u) %s)\n", file, (unsigned int)errNo, errMsg);
			break;
			
			case ERROR_NO_DELETE:
				printf("\nCould not delete file [%s] ((%u) %s)\nMake sure you have permission to delete the file, and that the file is not in use\n", file, (unsigned int)errNo, errMsg);
			break;
			
			case ERROR_NO_ALLOC:
				printf("\nFatal Error: Cannot allocate memory\n");
			
			default:
				printf("\nAn unknown error occurred\n");
			break;
		}
		if (errType != ERROR_NO_ALLOC)
			LocalFree(errMsg);
	}
	
	if (doExit == EXIT_EXIT)
		ExitA(files, errNo);
}

void DisplayErrorW(fileArrayW* files, DWORD errNo, int errType, const wchar_t* file, int doExit)
{
	if (errNo != ERROR_NO_ERROR)
	{
		wchar_t* errMsg;
		if (errType != ERROR_NO_ALLOC)
			errMsg = GetErrorMessageW(errNo);
		switch (errType)
		{
			case ERROR_TOO_LONG:
				wprintf(L"\nPath [%s] exceeds maximum path length of %u characters\n", file, MAX_PATH);
			break;
			
			case ERROR_NO_OPEN:
				wprintf(L"\nCannot open file/directory [%s] ((%u) %s)\n", file, (unsigned int)errNo, errMsg);
			break;
			
			case ERROR_INVALID_PATH:
				wprintf(L"\nPath [%s] is not a valid directory\n", file);
			break;
			
			case ERROR_NO_SIZE:
				wprintf(L"\nCannot get size of file [%s] ((%u) %s)\n", file, (unsigned int)errNo, errMsg);
			break;
			
			case ERROR_NO_DELETE:
				wprintf(L"\nCould not delete file [%s] ((%u) %s)\nMake sure you have permission to delete the file, and that the file is not in use\n", file, (unsigned int)errNo, errMsg);
			break;
			
			case ERROR_NO_ALLOC:
				wprintf(L"\nFatal Error: Cannot allocate memory\n");
			
			default:
				wprintf(L"\nAn unknown error occurred\n");
			break;
		}
		if (errType != ERROR_NO_ALLOC)
			LocalFree(errMsg);
	}
	
	if (doExit == EXIT_EXIT)
		ExitW(files, errNo);
}

void ExitA(fileArrayA* files, DWORD code)
{
	if (files->array != NULL)
	{
		for (size_t i = 0; i < files->elements; i++)
		{
			free(files->array[i].hash);
			free(files->array[i].fileInfo);
		}
		free(files->array);
	}
	exit(code);
}

void ExitW(fileArrayW* files, DWORD code)
{
	if (files->array != NULL)
	{
		for (size_t i = 0; i < files->elements; i++)
		{
			free(files->array[i].hash);
			free(files->array[i].fileInfo);
		}
		free(files->array);
	}
	exit(code);
}
