/************************************************************************************************
*  Filename: md5hash.c
*  Author: <kasskloff AT gmail.com>
*  Description: A (kind of fast) implementation of the MD5 hashing algorithm for little-endian machines
*  Uses optimizations from Colin Plumb's public domain code
*  
*  Permission is hereby granted, free of charge, to any person obtaining a copy
*  of this software and associated documentation files (the "Software"), to deal
*  in the Software without restriction, including without limitation the rights
*  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
*  copies of the Software, and to permit persons to whom the Software is
*  furnished to do so, subject to the following conditions:
*  
*  The above copyright notice and this permission notice shall be included in all
*  copies or substantial portions of the Software.
*  
*  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
*  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
*  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
*  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
*  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
*  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
*  SOFTWARE.
************************************************************************************************/

#include <stdlib.h>
#include <memory.h>
#include "md5hash.h"

#define rotl(x, n) ((x << n) | (x >> (32 - n)))

#define F1(x,y,z) (z ^ (x & (y ^ z)))
#define F2(x,y,z) (y ^ (z & (x ^ y)))
#define F3(x,y,z) (x ^ y ^ z)
#define F4(x,y,z) (y ^ (x | ~z))

#define R1(a,b,c,d,k,s,t) { a += F1(b,c,d) + m[k] + t; a = b + rotl(a,s); }
#define R2(a,b,c,d,k,s,t) { a += F2(b,c,d) + m[k] + t; a = b + rotl(a,s); }
#define R3(a,b,c,d,k,s,t) { a += F3(b,c,d) + m[k] + t; a = b + rotl(a,s); }
#define R4(a,b,c,d,k,s,t) { a += F4(b,c,d) + m[k] + t; a = b + rotl(a,s); }

void Transform_md5(MD5* md5, unsigned char* data);

static const unsigned char padding[64] = {
	0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
       0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
       0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

void Initialise_md5(MD5* md5)
{
	md5->a = 0x67452301;
	md5->b = 0xEFCDAB89;
	md5->c = 0x98BADCFE;
	md5->d = 0x10325476;
	md5->length[0] = 0;
	md5->length[1] = 0;
}

void Update_md5(MD5* md5, unsigned char* data, size_t length)
{
	if (length == 0)
		return;
	
	unsigned int left = md5->length[0] & 0x03F;
	unsigned int count = 64 - left;
	md5->length[0] += length;
	md5->length[0] = md5->length[0] & 0xFFFFFFFF;
	
	if (md5->length[0] < length)
		md5->length[1]++;
	
	if (left > 0 && length >= count)
	{
		memcpy((md5->data + left), data, count);
		Transform_md5(md5, md5->data);
		data += count;
		length -= count;
		left = 0;
	}
	
	while (length > 63)
	{
		Transform_md5(md5, data);
		data += 64;
		length -= 64;
	}
	
	if (length > 0)
		memcpy((md5->data + left), data, length);
}

void Finalise_md5(MD5* md5, unsigned char* hash)
{
	unsigned int highBits = (md5->length[0] >> 29) | (md5->length[1] << 3);
	unsigned int lowBits = md5->length[0] << 3;
	unsigned char messageLength[8];
	
	messageLength[     0] = lowBits;
	messageLength[ 0 + 1] = lowBits >> 8; 
	messageLength[ 0 + 2] = lowBits >> 16;
	messageLength[ 0 + 3] = lowBits >> 24;
	messageLength[     4] = highBits;
	messageLength[ 4 + 1] = highBits >> 8; 
	messageLength[ 4 + 2] = highBits >> 16;
	messageLength[ 4 + 3] = highBits >> 24;
	
	unsigned int count = md5->length[0] & 0x3F;
	unsigned int padAmount = ((count < 56) ? (56 - count) : (120 - count));
	Update_md5(md5, (unsigned char*)padding, padAmount);
	Update_md5(md5, messageLength, 8);
	
	hash[     0] = md5->a;
	hash[ 0 + 1] = md5->a >> 8; 
	hash[ 0 + 2] = md5->a >> 16;
	hash[ 0 + 3] = md5->a >> 24;
	hash[     4] = md5->b;
	hash[ 4 + 1] = md5->b >> 8; 
	hash[ 4 + 2] = md5->b >> 16;
	hash[ 4 + 3] = md5->b >> 24;
	hash[     8] = md5->c;
	hash[ 8 + 1] = md5->c >> 8; 
	hash[ 8 + 2] = md5->c >> 16;
	hash[ 8 + 3] = md5->c >> 24;
	hash[    12] = md5->d;
	hash[12 + 1] = md5->d >> 8; 
	hash[12 + 2] = md5->d >> 16;
	hash[12 + 3] = md5->d >> 24;
}

void Transform_md5(MD5* md5, unsigned char* data)
{
	unsigned int a = md5->a;
	unsigned int b = md5->b;
	unsigned int c = md5->c;
	unsigned int d = md5->d;
	unsigned int m[16];
	
	m[ 0] = (data[ 0]) | (data[ 0 + 1] << 8) | (data[ 0 + 2] << 16) | (data[ 0 + 3] << 24);
	m[ 1] = (data[ 4]) | (data[ 4 + 1] << 8) | (data[ 4 + 2] << 16) | (data[ 4 + 3] << 24);
	m[ 2] = (data[ 8]) | (data[ 8 + 1] << 8) | (data[ 8 + 2] << 16) | (data[ 8 + 3] << 24);
	m[ 3] = (data[12]) | (data[12 + 1] << 8) | (data[12 + 2] << 16) | (data[12 + 3] << 24);
	m[ 4] = (data[16]) | (data[16 + 1] << 8) | (data[16 + 2] << 16) | (data[16 + 3] << 24);
	m[ 5] = (data[20]) | (data[20 + 1] << 8) | (data[20 + 2] << 16) | (data[20 + 3] << 24);
	m[ 6] = (data[24]) | (data[24 + 1] << 8) | (data[24 + 2] << 16) | (data[24 + 3] << 24);
	m[ 7] = (data[28]) | (data[28 + 1] << 8) | (data[28 + 2] << 16) | (data[28 + 3] << 24);
	m[ 8] = (data[32]) | (data[32 + 1] << 8) | (data[32 + 2] << 16) | (data[32 + 3] << 24);
	m[ 9] = (data[36]) | (data[36 + 1] << 8) | (data[36 + 2] << 16) | (data[36 + 3] << 24);
	m[10] = (data[40]) | (data[40 + 1] << 8) | (data[40 + 2] << 16) | (data[40 + 3] << 24);
	m[11] = (data[44]) | (data[44 + 1] << 8) | (data[44 + 2] << 16) | (data[44 + 3] << 24);
	m[12] = (data[48]) | (data[48 + 1] << 8) | (data[48 + 2] << 16) | (data[48 + 3] << 24);
	m[13] = (data[52]) | (data[52 + 1] << 8) | (data[52 + 2] << 16) | (data[52 + 3] << 24);
	m[14] = (data[56]) | (data[56 + 1] << 8) | (data[56 + 2] << 16) | (data[56 + 3] << 24);
	m[15] = (data[60]) | (data[60 + 1] << 8) | (data[60 + 2] << 16) | (data[60 + 3] << 24);
	
	R1(a, b, c, d,  0,  7, 0xd76aa478);
	R1(d, a, b, c,  1, 12, 0xe8c7b756);
	R1(c, d, a, b,  2, 17, 0x242070db);
	R1(b, c, d, a,  3, 22, 0xc1bdceee);
	R1(a, b, c, d,  4,  7, 0xf57c0faf);
	R1(d, a, b, c,  5, 12, 0x4787c62a);
	R1(c, d, a, b,  6, 17, 0xa8304613);
	R1(b, c, d, a,  7, 22, 0xfd469501);
	R1(a, b, c, d,  8,  7, 0x698098d8);
	R1(d, a, b, c,  9, 12, 0x8b44f7af);	
	R1(c, d, a, b, 10, 17, 0xffff5bb1);
	R1(b, c, d, a, 11, 22, 0x895cd7be);
	R1(a, b, c, d, 12,  7, 0x6b901122);
	R1(d, a, b, c, 13, 12, 0xfd987193);
	R1(c, d, a, b, 14, 17, 0xa679438e);
	R1(b, c, d, a, 15, 22, 0x49b40821);

	R2(a, b, c, d,  1,  5, 0xf61e2562);
	R2(d, a, b, c,  6,  9, 0xc040b340);
	R2(c, d, a, b, 11, 14, 0x265e5a51);
	R2(b, c, d, a,  0, 20, 0xe9b6c7aa);
	R2(a, b, c, d,  5,  5, 0xd62f105d);
	R2(d, a, b, c, 10,  9, 0x02441453);
	R2(c, d, a, b, 15, 14, 0xd8a1e681);
	R2(b, c, d, a,  4, 20, 0xe7d3fbc8);
	R2(a, b, c, d,  9,  5, 0x21e1cde6);
	R2(d, a, b, c, 14,  9, 0xc33707d6);
	R2(c, d, a, b,  3, 14, 0xf4d50d87);
	R2(b, c, d, a,  8, 20, 0x455a14ed);
	R2(a, b, c, d, 13,  5, 0xa9e3e905);
	R2(d, a, b, c,  2,  9, 0xfcefa3f8);
	R2(c, d, a, b,  7, 14, 0x676f02d9);
	R2(b, c, d, a, 12, 20, 0x8d2a4c8a);

	R3(a, b, c, d,  5,  4, 0xfffa3942);
	R3(d, a, b, c,  8, 11, 0x8771f681);
	R3(c, d, a, b, 11, 16, 0x6d9d6122);
	R3(b, c, d, a, 14, 23, 0xfde5380c);
	R3(a, b, c, d,  1,  4, 0xa4beea44);
	R3(d, a, b, c,  4, 11, 0x4bdecfa9);
	R3(c, d, a, b,  7, 16, 0xf6bb4b60);
	R3(b, c, d, a, 10, 23, 0xbebfbc70);
	R3(a, b, c, d, 13,  4, 0x289b7ec6);
	R3(d, a, b, c,  0, 11, 0xeaa127fa);
	R3(c, d, a, b,  3, 16, 0xd4ef3085);
	R3(b, c, d, a,  6, 23, 0x04881d05);
	R3(a, b, c, d,  9,  4, 0xd9d4d039);
	R3(d, a, b, c, 12, 11, 0xe6db99e5);
	R3(c, d, a, b, 15, 16, 0x1fa27cf8);
	R3(b, c, d, a,  2, 23, 0xc4ac5665);

	R4(a, b, c, d,  0,  6, 0xf4292244);
	R4(d, a, b, c,  7, 10, 0x432aff97);
	R4(c, d, a, b, 14, 15, 0xab9423a7);
	R4(b, c, d, a,  5, 21, 0xfc93a039);
	R4(a, b, c, d, 12,  6, 0x655b59c3);
	R4(d, a, b, c,  3, 10, 0x8f0ccc92);
	R4(c, d, a, b, 10, 15, 0xffeff47d);
	R4(b, c, d, a,  1, 21, 0x85845dd1);
	R4(a, b, c, d,  8,  6, 0x6fa87e4f);
	R4(d, a, b, c, 15, 10, 0xfe2ce6e0);
	R4(c, d, a, b,  6, 15, 0xa3014314);
	R4(b, c, d, a, 13, 21, 0x4e0811a1);
	R4(a, b, c, d,  4,  6, 0xf7537e82);
	R4(d, a, b, c, 11, 10, 0xbd3af235);
	R4(c, d, a, b,  2, 15, 0x2ad7d2bb);
	R4(b, c, d, a,  9, 21, 0xeb86d391);
	
	md5->a += a;
	md5->b += b;
	md5->c += c;
	md5->d += d;
}
